"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_nat_gateway
hub.exec.boto3.client.ec2.delete_nat_gateway
hub.exec.boto3.client.ec2.describe_nat_gateways
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": ["aws.ec2.subnet.present"],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    subnet_id: str,
    resource_id: str = None,
    connectivity_type: str = "public",
    client_token: str = None,
    allocation_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a NAT gateway in the specified subnet. This action creates a network interface in the specified subnet
    with a private IP address from the IP address range of the subnet. You can create either a public NAT gateway or
    a private NAT gateway. With a public NAT gateway, internet-bound traffic from a private subnet can be routed to
    the NAT gateway, so that instances in a private subnet can connect to the internet. With a private NAT gateway,
    private communication is routed across VPCs and on-premises networks through a transit gateway or virtual
    private gateway. Common use cases include running large workloads behind a small pool of allowlisted IPv4
    addresses, preserving private IPv4 addresses, and communicating between overlapping networks. For more
    information, see NAT gateways in the Amazon Virtual Private Cloud User Guide.

    Args:
        name(Text): An Idem name to identify the NAT gateway resource.
        subnet_id(Text): The subnet in which to create the NAT gateway.
        resource_id(Text, optional): AWS Internet Gateway ID.
        client_token(Text, optional): Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. Constraint: Maximum 64 ASCII characters. This field is autopopulated if not provided.
        connectivity_type(Text, optional): Indicates whether the NAT gateway supports public or private connectivity. The default is public connectivity.
        allocation_id(Text, optional): [Public NAT gateway only]. The allocation ID of the Elastic IP address that's associated with the NAT gateway.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the NAT gateway resource. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS IAM Policy.
            * create (Dict): Timeout configuration for creating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
            * update(Dict, optional): Timeout configuration for updating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [nat_gateway-name]:
          aws.ec2.nat_gateway.present:
          - subnet_id: 'string'
          - resource_id: 'string'
          - client_token: 'string'
          - connectivity_type: 'string'
          - allocation_id: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'
          - timeout:
            create:
              delay: 'integer'
              max_attempts: 'integer'
            update:
              delay: 'integer'
              max_attempts: 'integer'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-nat-gateway:
              aws.ec2.nat_gateway.present:
                - subnet_id: subnet-026542dd856a97e15
                - client_token: nat-0e7969ddf4c38831f
                - resource_id: nat-0e7969ddf4c38831f
                - connectivity_type: private
                - tags:
                  - Key: Name
                    Value: test-nat-gateway
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before = await hub.exec.boto3.client.ec2.describe_nat_gateways(
            ctx, NatGatewayIds=[resource_id]
        )
    if before and before["result"] and before["ret"].get("NatGateways"):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_nat_gateway_to_present(
            raw_resource=before["ret"]["NatGateways"][0], idem_resource_name=name
        )
        old_tags = result["old_state"].get("tags")
        plan_state = copy.deepcopy(result["old_state"])
        # check if tags need to be updated
        if tags is not None and tags != old_tags:
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=old_tags,
                new_tags=tags,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
            elif ctx.get("test", False):
                plan_state["tags"] = update_ret["ret"]
                result["comment"] = (
                    f"Would update tags for aws.ec2.nat_gateway {name}",
                )
            else:
                resource_updated = True
                result["comment"] = result["comment"] + update_ret["comment"]
        if resource_updated:
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("update") if timeout else None,
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "ec2",
                    "nat_gateway_available",
                    NatGatewayIds=[resource_id],
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
        elif not ctx.get("test", False):
            result["comment"] = (
                f"aws.ec2.nat_gateway '{name}' has no property need to be updated",
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "allocation_id": allocation_id,
                    "client_token": client_token,
                    "subnet_id": subnet_id,
                    "name": name,
                    "tags": tags,
                    "connectivity_type": connectivity_type,
                },
            )
            result["comment"] = (f"Would create aws.ec2.nat_gateway '{name}'",)
            return result
        else:
            ret = await hub.exec.boto3.client.ec2.create_nat_gateway(
                ctx,
                AllocationId=allocation_id,
                ClientToken=client_token,
                SubnetId=subnet_id,
                TagSpecifications=[
                    {
                        "ResourceType": "natgateway",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags is not None
                else None,
                ConnectivityType=connectivity_type,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = ret["ret"]["NatGateway"]["NatGatewayId"]
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "ec2",
                    "nat_gateway_available",
                    NatGatewayIds=[resource_id],
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            if result["result"]:
                result["comment"] = (f"Created '{resource_id}'",)
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.ec2.describe_nat_gateways(
                ctx, NatGatewayIds=[resource_id]
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_nat_gateway_to_present(
                raw_resource=after["ret"]["NatGateways"][0], idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, timeout: Dict = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified NAT gateway. Deleting a public NAT gateway disassociates its Elastic IP address, but does
    not release the address from your account. Deleting a NAT gateway does not delete any NAT gateway routes in your
    route tables.

    Args:
        name(Text): An Idem name to identify the NAT gateway resource.
        resource_id(Text, optional): The AWS ID of the nat gateway. Idem automatically considers this resource being absent
         if this field is not specified.
        timeout(Dict, optional): Timeout configuration for deletion of AWS Nat Gateway.
            * delete (Dict) -- Timeout configuration for deletion of a Nat Gateway
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [nat_gateway-name]:
          aws.ec2.nat_gateway.absent:
          - name: 'string'
          - resource_id: 'string'
          - timeout:
              delete:
                delay: 'integer'
                max_attempts: 'integer'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-nat-gateway:
                aws.ec2.nat_gateway.absent:
                  - name: my-nat-gateway
                  - resource_id: nat-0e7969ddf4c38831f
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.nat_gateway", name=name
        )
        return result
    before = await hub.exec.boto3.client.ec2.describe_nat_gateways(
        ctx, NatGatewayIds=[resource_id]
    )
    if not before["result"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.nat_gateway", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_nat_gateway_to_present(
            raw_resource=before, idem_resource_name=name
        )
        result["comment"] = (f"Would delete aws.ec2.nat_gateway {name}",)
        return result
    else:
        try:
            if before["ret"]["NatGateways"][0]["State"] == "deleted":
                result["comment"] = (
                    f"aws.ec2.nat_gateway '{name}' is in deleted state.",
                )
            else:
                result[
                    "old_state"
                ] = hub.tool.aws.ec2.conversion_utils.convert_raw_nat_gateway_to_present(
                    raw_resource=before["ret"]["NatGateways"][0],
                    idem_resource_name=name,
                )
                ret = await hub.exec.boto3.client.ec2.delete_nat_gateway(
                    ctx, NatGatewayId=resource_id
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=15,
                    default_max_attempts=40,
                    timeout_config=timeout.get("delete") if timeout else None,
                )
                acceptors = [
                    {
                        "matcher": "pathAll",
                        "expected": "deleted",
                        "state": "success",
                        "argument": "NatGateways[].State",
                    },
                    {
                        "matcher": "pathAll",
                        "expected": "deleting",
                        "state": "retry",
                        "argument": "NatGateways[].State",
                    },
                ]
                cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                    name="NatGatewayDeleted",
                    operation="DescribeNatGateways",
                    argument=["NatGateways[].State"],
                    acceptors=acceptors,
                    client=hub.tool.boto3.client.get_client(ctx, "ec2"),
                )
                await hub.tool.boto3.client.wait(
                    ctx,
                    "ec2",
                    "NatGatewayDeleted",
                    cluster_waiter,
                    NatGatewayIds=[resource_id],
                    WaiterConfig=waiter_config,
                )
                result["comment"] = (f"Deleted '{name}'",)

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] = result["comment"] + (
                f"{e.__class__.__name__}: {e}",
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your NAT gateways.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.nat_gateway
    """
    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_nat_gateways(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe nat_gateway {ret['comment']}")
        return result

    for nat_gateway in ret["ret"]["NatGateways"]:
        nat_gateway_id = nat_gateway.get("NatGatewayId")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_nat_gateway_to_present(
                raw_resource=nat_gateway, idem_resource_name=nat_gateway_id
            )
        )
        result[nat_gateway_id] = {
            "aws.ec2.nat_gateway.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



import copy
from typing import Any
from typing import Dict


async def get(hub, ctx, name: str, resource_id: str, rest_api_id: str):
    """
    Get an API Gateway Resource from AWS.
    Args:
        hub:
        ctx:
        name(string): An Idem name of the API Gateway Resource.
        resource_id(string): AWS Resource id. Defaults to None.
        rest_api_id(string): AWS rest_api id of the associated RestApi.

    """

    result = dict(comment=[], ret=None, result=True)
    before = await hub.exec.boto3.client.apigateway.get_resource(
        ctx, resourceId=resource_id, restApiId=rest_api_id
    )

    if not before["result"]:
        if "NotFoundException" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.api.gateway", name=name
                )
            )
            result["comment"] += list(before["comment"])
            return result
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if before["ret"]:
        result["ret"] = before["ret"]

    return result


async def update_resource(
    hub, ctx, old_state: Dict[str, Any], update_parameters: Dict[str, any]
):
    r"""

    Updates an API Gateway Resource

    Args:
        hub:
        ctx:
        old_state(Dict): Previous state of resource
        update_parameters(Dict): Parameters from SLS File

    Returns:
        Dict[str, Any]

    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False

    for key, value in update_parameters.items():
        if key == "parent_id":
            key_final = "parentId"
        else:
            key_final = "pathPart"

        if old_state.get(key) and value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, list)
                    else ",".join(value),
                }
            )
            updated = True

        elif old_state.get(key) is None:
            patch_ops.append(
                {
                    "op": "add",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, list)
                    else ",".join(value),
                }
            )

            new_state[key] = value
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.resource", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.resource", name=old_state["name"]
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_resource(
            ctx,
            restApiId=old_state["rest_api_id"],
            resourceId=old_state["resource_id"],
            patchOperations=patch_ops,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.resource", name=old_state["name"]
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.resource", name=old_state["name"]
        )

    return result



from typing import Any
from typing import Dict
from typing import List


async def search_raw(
    hub,
    ctx,
    name,
    resource_id: str = None,
    availability_zone: str = None,
    availability_zone_id: str = None,
    cidr_block: str = None,
    default_for_az: str = None,
    filters: List = None,
    ipv6_cidr_block: str = None,
    status: str = None,
    vpc_id: str = None,
    tags: List or Dict = None,
) -> Dict:
    """
    Fetch one or more subnets from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS subnet id to identify the resource.
        availability_zone(string, optional): The Availability Zone for the subnet.
        availability_zone_id(string, optional): The ID of the Availability Zone for the subnet.
        cidr_block(string, optional): The IPv4 CIDR block of the subnet. The CIDR block you specify must exactly match
         the subnet's CIDR block for information to be returned for the subnet.
        default_for_az(string, optional): Indicate whether the subnet is the default subnet in the Availability Zone.
        filters(List, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_subnets
        ipv6_cidr_block(string, optional): An IPv6 CIDR block associated with the subnet.
        status(string, optional): The state of the subnet (pending | available ).
        vpc_id(string, optional): The ID of the VPC for the subnet.
        tags(List, optional): The list of tags to filter by. For example, to find all resources that have a tag with the key
            "Owner" and the value "TeamA" , specify "tag:Owner" for the Dict key and "TeamA" for the Dict value.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    filter_params = {
        "availability-zone": availability_zone,
        "availability-zone-id": availability_zone_id,
        "cidr-block": cidr_block,
        "default-for-az": default_for_az,
        "ipv6-cidr-block-association.ipv6-cidr-block": ipv6_cidr_block,
        "state": status,
        "vpc-id": vpc_id,
    }

    for filter_param, value in filter_params.items():
        if value is not None:
            boto3_filter.append({"Name": filter_param, "Values": [value]})
    if tags is not None:
        for tag in tags:
            boto3_filter.append({"Name": f"tag:{tag['Key']}", "Values": [tag["Value"]]})

    ret = await hub.exec.boto3.client.ec2.describe_subnets(
        ctx,
        Filters=boto3_filter,
        SubnetIds=[resource_id] if resource_id else None,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update_ipv6_cidr_blocks(
    hub,
    ctx,
    subnet_id: str,
    old_ipv6_cidr_block: Dict[str, Any],
    new_ipv6_cidr_block: Dict[str, Any],
):
    """
    Update associated ipv6 cidr block of a subnet. This function supports associating an ipv6 cidr block, or updating
     the existing(old) ipv6 cidr block to the new ipv6 cidr block. Disassociating an ipv6 cidr block is not supported,
     due to how an sls file works currently. If ipv6_cidr_block parameter is left as blank, Idem-aws will do no-op
     on the subnet's ipv6 cidr block. To disassociate an ipv6 cidr block, a user will have to delete the subnet and
     re-create it without the ipv6 cidr block.

    Args:
        hub:
        ctx:
        subnet_id: The AWS resource id of the existing subnet
        old_ipv6_cidr_block: The ipv6 cidr block on the existing vpc
        new_ipv6_cidr_block: The expected ipv6 cidr block on the existing vpc

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    if old_ipv6_cidr_block is None and new_ipv6_cidr_block:
        if ctx.get("test", False):
            result["ret"] = {
                "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
            }
            return result
        else:
            ret = await hub.exec.boto3.client.ec2.associate_subnet_cidr_block(
                ctx,
                SubnetId=subnet_id,
                Ipv6CidrBlock=new_ipv6_cidr_block.get("Ipv6CidrBlock"),
            )
            result["result"] = ret["result"]
            if result["result"]:
                hub.log.info(
                    f"Add subnet {subnet_id} ipv6 cidr block {new_ipv6_cidr_block.get('Ipv6CidrBlock')}"
                )
                result["ret"] = {
                    "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
                }
            else:
                result["comment"] = ret["comment"]
            return result
    elif old_ipv6_cidr_block and new_ipv6_cidr_block:
        if old_ipv6_cidr_block.get("Ipv6CidrBlock") != new_ipv6_cidr_block.get(
            "Ipv6CidrBlock"
        ):
            if ctx.get("test", False):
                result["ret"] = {
                    "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
                }
                return result
            else:
                ret = await hub.exec.boto3.client.ec2.disassociate_subnet_cidr_block(
                    ctx, AssociationId=old_ipv6_cidr_block.get("AssociationId")
                )
                if not ret.get("result"):
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                ret = await hub.exec.boto3.client.ec2.associate_subnet_cidr_block(
                    ctx,
                    SubnetId=subnet_id,
                    Ipv6CidrBlock=new_ipv6_cidr_block.get("Ipv6CidrBlock"),
                )
                result["result"] = ret["result"]
                if result["result"]:
                    hub.log.info(
                        f"Update subnet {subnet_id} ipv6 cidr block from {old_ipv6_cidr_block.get('Ipv6CidrBlock')}"
                        f" to {new_ipv6_cidr_block.get('Ipv6CidrBlock')}"
                    )
                    result["ret"] = {
                        "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
                    }
                    return result
                else:
                    result["comment"] = ret["comment"]
    return result


async def update_subnet_attributes(
    hub,
    ctx,
    before: Dict[str, Any],
    resource_id: str,
    map_public_ip_on_launch: bool,
    assign_ipv6_address_on_creation: bool,
    map_customer_owned_ip_on_launch: bool,
    customer_owned_ipv4_pool: str,
    enable_dns_64: bool,
    private_dns_name_options_on_launch: Dict[str, Any],
    enable_lni_at_device_index: int,
    disable_lni_at_device_index: bool,
):
    """
    Updates the Subnet Attributes
    Args:
        before: existing resource
        resource_id (Text): AWS Subnet ID
        map_public_ip_on_launch (boolean): Indicates whether instances launched in this subnet receive a public IPv4 address.
        assign_ipv6_address_on_creation (boolean): Specify true to indicate that network interfaces created in the specified subnet should be assigned an IPv6 address.
        map_customer_owned_ip_on_launch (boolean): Specify true to indicate that network interfaces attached to instances created in the specified subnet should be assigned a customer-owned IPv4 address.
        customer_owned_ipv4_pool (Text): The customer-owned IPv4 address pool associated with the subnet.
        enable_dns_64 (boolean): Indicates whether DNS queries made to the Amazon-provided DNS Resolver in this subnet should return synthetic IPv6 addresses for IPv4-only destinations.
        private_dns_name_options_on_launch (Dict): The type of hostnames to assign to instances in the subnet at launch.
        enable_lni_at_device_index (int): Indicates the device position for local network interfaces in this subnet.
        disable_lni_at_device_index (boolean): Specify true to indicate that local network interfaces at the current position should be disabled.


    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None|Updated Attributes name and value}

    """
    result = dict(comment=(), result=True, ret={})

    if (
        before.get("private_dns_name_options_on_launch") is not None
        and private_dns_name_options_on_launch is not None
    ):
        update_ret = await modify_subnet_attributes_dns_options(
            hub,
            ctx,
            resource_id,
            private_dns_name_options_on_launch,
            before["private_dns_name_options_on_launch"],
        )
        result["ret"] = update_ret["ret"]
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = result["result"] and update_ret["result"]

    resource_attributes = {
        "assign_ipv6_address_on_creation": assign_ipv6_address_on_creation,
        "map_public_ip_on_launch": map_public_ip_on_launch,
        "customer_owned_ipv4_pool": customer_owned_ipv4_pool,
        "map_customer_owned_ip_on_launch": map_customer_owned_ip_on_launch,
        "enable_dns_64": enable_dns_64,
        "enable_lni_at_device_index": enable_lni_at_device_index,
        "disable_lni_at_device_index": disable_lni_at_device_index,
    }

    for key, value in resource_attributes.items():
        if key in before:
            if (value is not None) and value != before[key]:
                update_ret = await modify_subnet_attribute(
                    hub, ctx, resource_id, key, resource_attributes[key]
                )
                result["ret"].update(update_ret["ret"])
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]
    return result


async def modify_subnet_attributes_dns_options(
    hub,
    ctx,
    resource_id: str,
    private_dns_name_options: Dict[str, str],
    before_dns_name_options: Dict[str, str],
):
    """
    Modify the subnet attributes for dns name options
    Args:
        before_dns_name_options (Dict): The type of hostnames to assign to instance in the subnet at launch.
        resource_id (Text): AWS Subnet ID
        private_dns_name_options (Dict): The type of hostnames to assign to instance in the subnet at launch.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None|Updated Attributes name and value}

    """
    result = dict(comment=(), result=True, ret={})

    host_name_type = private_dns_name_options.get("HostnameType")
    enable_resource_name_dns_a_record = private_dns_name_options.get(
        "EnableResourceNameDnsARecord"
    )
    enable_resource_name_dns_aaaa_record = private_dns_name_options.get(
        "EnableResourceNameDnsAAAARecord"
    )

    # Added key value for private_dns_options because for update call, parameter names are different.
    private_dns_keys_values = {
        "HostnameType": "PrivateDnsHostnameTypeOnLaunch",
        "EnableResourceNameDnsARecord": "EnableResourceNameDnsARecordOnLaunch",
        "EnableResourceNameDnsAAAARecord": "EnableResourceNameDnsAAAARecordOnLaunch",
    }

    dns_key_values = {
        "HostnameType": host_name_type,
        "EnableResourceNameDnsARecord": enable_resource_name_dns_a_record,
        "EnableResourceNameDnsAAAARecord": enable_resource_name_dns_aaaa_record,
    }

    # This is written separately because all the attributes in dns_name_options dict needs to be updated.
    for key, value in dns_key_values.items():
        if key in before_dns_name_options:
            if (value is not None) and value != before_dns_name_options.get(key):
                update_ret = await modify_subnet_attribute(
                    hub, ctx, resource_id, private_dns_keys_values[key], value
                )
                result["ret"].update(update_ret["ret"])
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]

    return result


async def modify_subnet_attribute(
    hub, ctx, resource_id: str, attr_key: str, value: str
):
    result = dict(comment=(), result=True, ret={})
    update_payload = {}
    str_objects = [
        "customer_owned_ipv4_pool",
        "private_dns_name_options_on_launch",
        "enable_lni_at_device_index",
        "PrivateDnsHostnameTypeOnLaunch",
    ]

    resource_parameters = {
        "map_public_ip_on_launch": "MapPublicIpOnLaunch",
        "assign_ipv6_address_on_creation": "AssignIpv6AddressOnCreation",
        "map_customer_owned_ip_on_launch": "MapCustomerOwnedIpOnLaunch",
        "customer_owned_ipv4_pool": "CustomerOwnedIpv4Pool",
        "enable_dns_64": "EnableDns64",
        "private_dns_name_options_on_launch": "PrivateDnsNameOptionsOnLaunch",
        "PrivateDnsHostnameTypeOnLaunch": "PrivateDnsHostnameTypeOnLaunch",
        "EnableResourceNameDnsARecordOnLaunch": "EnableResourceNameDnsARecordOnLaunch",
        "EnableResourceNameDnsAAAARecordOnLaunch": "EnableResourceNameDnsAAAARecordOnLaunch",
        "enable_lni_at_device_index": "EnableLniAtDeviceIndex",
    }

    attr_value = value
    if not ctx.get("test", False):
        if attr_key not in str_objects:
            attr_value = {"Value": value}
        update_payload[resource_parameters[attr_key]] = attr_value
        update_ret = await hub.exec.boto3.client.ec2.modify_subnet_attribute(
            ctx, SubnetId=resource_id, **update_payload
        )
        if not update_ret["result"]:
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = False
            return result

    result["ret"] = {attr_key: value}
    result["comment"] += (f"Update {attr_key}",)
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.attach_user_policy
hub.exec.boto3.client.iam.detach_user_policy
hub.exec.boto3.client.iam.list_user_policies
"""
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub, ctx, name: str, user_name: str, policy_arn: str
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Attaches the specified managed policy to the specified user.
    You use this operation to attach a managed policy to a user. To embed an inline policy in a user, use PutUserPolicy .

    Args:
        hub:
        ctx:
        name(Text): The name (friendly name, not ARN) of the IAM user to attach the policy to.
        user_name(Text): The name (friendly name, not ARN) of the IAM user to detach the policy from.
        This parameter allows (through its regex pattern ) a string of characters consisting of upper and lowercase alphanumeric characters with no spaces. You can also include any of the following characters: _+=,.@-
        policy_arn(Text): The Amazon Resource Name (ARN) of the IAM policy you want to attach.

    Request Syntax:
        [iam-attach-user-policy-name]:
          aws.iam.user_policy_attachment.present:
          - user_name: 'string'
          - policy_arn: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls
        serverless-arn:aws:iam::aws:policy/AdministratorAccess:
          aws.iam.user_policy_attachment.present:
          - user_name: serverless
          - policy_arn: arn:aws:iam::aws:policy/AdministratorAccess

    """
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    try:
        before_ret = await hub.exec.aws.iam.user_policy_attachment.get_policy_if_attached_to_user(
            ctx, user_name=user_name, policy_arn=policy_arn
        )
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False
        return result
    if ctx.get("test", False):
        if before_ret["result"]:
            result["comment"] = f"aws.iam.user_policy_attachment {name} already exists"
        else:
            result["comment"] = f"Would attach aws.iam.user_policy_attachment {name}"
        return result

    if before_ret["result"] is False and before_ret["comment"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if before_ret["result"]:
        result["old_state"] = before_ret["ret"]
        result["new_state"] = before_ret["ret"]
        result["comment"] = f"'{name}' already exists"
    else:
        try:
            ret = await hub.exec.boto3.client.iam.attach_user_policy(
                ctx,
                UserName=user_name,
                PolicyArn=policy_arn,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Attached '{name}'"
            result["new_state"] = {"PolicyArn": policy_arn}
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, user_name: str, policy_arn: str
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Removes the specified managed policy from the specified user.
    A user can also have inline policies embedded with it. To delete an inline policy, use DeleteUserPolicy

    Args:
        name(Text): The name (friendly name, not ARN) of the IAM user to attach the policy to.
        user_name(Text): The name (friendly name, not ARN) of the IAM user to detach the policy from.
        This parameter allows (through its regex pattern ) a string of characters consisting of upper and lowercase alphanumeric characters with no spaces. You can also include any of the following characters: _+=,.@-
        policy_arn(Text): The Amazon Resource Name (ARN) of the IAM policy you want to attach.

    Request Syntax:
        [iam-user-policy-name]:
          aws.iam.user_policy.absent:
          - user_name: 'string'
          - policy_arn: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

        serverless-arn:aws:iam::aws:policy/AdministratorAccess:
          aws.iam.user_policy_attachment.absent:
          - user_name: serverless
          - policy_arn: arn:aws:iam::aws:policy/AdministratorAccess
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    try:
        before_ret = await hub.exec.aws.iam.user_policy_attachment.get_policy_if_attached_to_user(
            ctx, user_name=user_name, policy_arn=policy_arn
        )
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False
        return result

    if before_ret["result"] is False and before_ret["comment"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result
    if not before_ret["result"]:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would detach aws.iam.user_policy_attachment {name}"
        return result
    else:
        try:
            result["old_state"] = before_ret["ret"]
            ret = await hub.exec.boto3.client.iam.detach_user_policy(
                ctx, UserName=user_name, PolicyArn=policy_arn
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"Detached '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists all managed policies that are attached to the specified IAM user. Lists all managed policies that are attached to the specified IAM user.



    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws_auto.iam.user_policy_attachment
    """

    result = {}
    # Fetch all users and then policies for each of them
    users = await hub.exec.boto3.client.iam.list_users(ctx)
    if not users["result"]:
        hub.log.debug(f"Could not describe user {users['comment']}")
        return {}
    user_names = [user.get("UserName") for user in users["ret"]["Users"]]
    for user_name in user_names:
        ret_attached_policies = (
            await hub.exec.boto3.client.iam.list_attached_user_policies(
                ctx=ctx, UserName=user_name
            )
        )
        if not ret_attached_policies["result"]:
            hub.log.warning(
                f"Could not get attached policy list with user {user_name} with error"
                f" {ret_attached_policies['comment']} . Describe will skip this user and continue."
            )
        else:
            resources = ret_attached_policies["ret"].get("AttachedPolicies")
            for resource in resources:
                resource_name = f"{user_name}-{resource.get('PolicyArn')}"
                resource_translated = list()
                resource_translated.append({"user_name": user_name})
                resource_translated.append({"policy_arn": resource.get("PolicyArn")})
                result[resource_name] = {
                    "aws.iam.user_policy_attachment.present": resource_translated
                }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.neptune.create_db_subnet_group
hub.exec.boto3.client.neptune.delete_db_subnet_group
hub.exec.boto3.client.neptune.describe_db_subnet_groups
hub.exec.boto3.client.neptune.modify_db_subnet_group
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_subnet_group_description: str,
    subnet_ids: List,
    resource_id: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new DB subnet group. DB subnet groups must contain at least one subnet in at least two AZs in the
    Amazon Region.

    Args:
        name(Text): An Idem name of the resource.
        db_subnet_group_description(Text): The description for the DB subnet group.
        subnet_ids(List): The EC2 Subnet IDs for the DB subnet group.
        resource_id(Text, optional): AWS Neptune DB Subnet Group Name. Defaults to None.
        tags(List, optional): The tags to be assigned to the new DB subnet group. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.neptune.db_subnet_group.present:
                - name: value
                - db_subnet_group_description: value
                - subnet_ids:
                  - subnet-123124324
                  - subnet-123124322
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.neptune.describe_db_subnet_groups(
            ctx, DBSubnetGroupName=resource_id
        )
        if not before["result"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

    if before and before["ret"].get("DBSubnetGroups"):
        resource_arn = before["ret"].get("DBSubnetGroups")[0].get("DBSubnetGroupArn")
        old_tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not old_tags["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + old_tags["comment"]
            return result
        old_tags = old_tags["ret"]
        resource_translated = (
            hub.tool.aws.neptune.db_subnet_group.convert_raw_db_subnet_group_to_present(
                before["ret"].get("DBSubnetGroups")[0],
                idem_resource_name=resource_id,
                tags=old_tags,
            )
        )
        result["old_state"] = resource_translated
        plan_state = copy.deepcopy(result["old_state"])

        old_subnets = result["old_state"].get("subnet_ids", [])
        update_parameters = {"DBSubnetGroupName": name, "SubnetIds": old_subnets}
        if subnet_ids is not None and set(old_subnets) != set(subnet_ids):
            resource_updated = True
            update_parameters["SubnetIds"] = subnet_ids

        old_description = result["old_state"].get("db_subnet_group_description")
        if (
            db_subnet_group_description is not None
            and old_description != db_subnet_group_description
        ):
            resource_updated = True
            update_parameters["DBSubnetGroupDescription"] = db_subnet_group_description

        if resource_updated:
            if ctx.get("test", False):
                plan_state["subnet_ids"] = subnet_ids
                plan_state["db_subnet_group_description"] = db_subnet_group_description
            else:
                # SubnetIds and DBSubnetGroupName are required parameters for this function
                # DBSubnetGroupDescription is not required however
                ret = await hub.exec.boto3.client.neptune.modify_db_subnet_group(
                    ctx, **update_parameters
                )
                result["result"] = ret["result"]
                result["comment"] = result["comment"] + ret["comment"]
                if not result["result"]:
                    return result
        if (
            tags is not None
            and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                tags, old_tags
            )
        ):
            resource_updated = True
            update_ret = await hub.exec.aws.neptune.tag.update_tags(
                ctx=ctx,
                resource_arn=resource_arn,
                old_tags=old_tags,
                new_tags=tags,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                hub.log.debug(
                    f"Failed updating tags for aws.neptune.db_subnet_group '{name}'"
                )
                return result
            result["comment"] = result["comment"] + update_ret["comment"]
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["tags"] = update_ret["ret"].get("tags")

        if ctx.get("test", False):
            if resource_updated:
                result["new_state"] = plan_state
                result["comment"] = (
                    f"Would update aws.neptune.db_subnet_group '{name}'",
                )
            else:
                result["comment"] = (
                    f"aws.neptune.db_subnet_group '{name}' already exists",
                )
                result["new_state"] = copy.deepcopy(result["old_state"])
            return result
        if not resource_updated:
            result["comment"] = (
                f"aws.neptune.db_subnet_group '{name}' already exists",
            )
        else:
            result["comment"] = result["comment"] + (
                f"Updated aws.neptune.db_subnet_group '{name}'",
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "db_subnet_group_description": db_subnet_group_description,
                    "resource_id": resource_id,
                    "subnet_ids": subnet_ids,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.neptune.db_subnet_group", name=name
            )
            return result
        ret = await hub.exec.boto3.client.neptune.create_db_subnet_group(
            ctx,
            DBSubnetGroupName=name,
            DBSubnetGroupDescription=db_subnet_group_description,
            SubnetIds=subnet_ids,
            Tags=tags,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["DBSubnetGroup"]["DBSubnetGroupName"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.neptune.db_subnet_group", name=name
        )
    try:
        if (not before) or resource_updated:
            after = await hub.exec.boto3.client.neptune.describe_db_subnet_groups(
                ctx, DBSubnetGroupName=resource_id
            )
            if not (
                after["result"] and after["ret"] and after["ret"].get("DBSubnetGroups")
            ):
                result["result"] = False
                result["comment"] = result["comment"] + after["comment"]
                return result
            resource_arn = after["ret"].get("DBSubnetGroups")[0].get("DBSubnetGroupArn")
            tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
                ctx, resource_arn=resource_arn
            )
            if not tags["result"]:
                result["result"] = False
                result["comment"] = result["comment"] + tags["comment"]
                return result
            tags = tags["ret"]
            resource_translated = hub.tool.aws.neptune.db_subnet_group.convert_raw_db_subnet_group_to_present(
                after["ret"].get("DBSubnetGroups")[0],
                idem_resource_name=name,
                tags=tags,
            )
            result["new_state"] = resource_translated
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a DB subnet group.  The specified database subnet group must not be associated with any DB instances.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, Optional): AWS Neptune DB Subnet Group Name. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.neptune.db_subnet_group.absent:
                - name: value
                - resource_id: value
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.neptune.db_subnet_group", name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.boto3.client.neptune.describe_db_subnet_groups(
        ctx, DBSubnetGroupName=resource_id
    )
    db_subnet_groups = (
        before["ret"].get("DBSubnetGroups") if before.get("ret") else None
    )

    if not (before["result"] and db_subnet_groups):
        # This condition means an error other than the resource not being found has happened
        if not ("DBSubnetGroupNotFoundFault" in str(before["comment"])):
            result["comment"] = before["comment"]
        return result

    resource_arn = db_subnet_groups[0].get("DBSubnetGroupArn")
    tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
        ctx, resource_arn=resource_arn
    )
    if not tags["result"]:
        result["result"] = False
        result["comment"] = tags["comment"]
        return result
    tags = tags["ret"]
    result[
        "old_state"
    ] = hub.tool.aws.neptune.db_subnet_group.convert_raw_db_subnet_group_to_present(
        db_subnet_groups[0], idem_resource_name=resource_id, tags=tags
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.neptune.db_subnet_group", name=name
        )
    else:
        ret = await hub.exec.boto3.client.neptune.delete_db_subnet_group(
            ctx, DBSubnetGroupName=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.neptune.db_subnet_group", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of DBSubnetGroup descriptions.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.neptune.db_subnet_group
    """

    result = {}
    ret = await hub.exec.boto3.client.neptune.describe_db_subnet_groups(ctx)

    if not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.neptune.db_subnet_group {ret['comment']}"
        )
    else:
        for db_subnet_group in ret["ret"]["DBSubnetGroups"]:
            resource_id = db_subnet_group.get("DBSubnetGroupName")
            resource_arn = db_subnet_group.get("DBSubnetGroupArn")
            tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
                ctx, resource_arn=resource_arn
            )
            if not tags["result"]:
                # if something goes wrong fetching the tags (not if the tags are None that is a normal path)
                hub.log.warning(
                    f"Failed listing tags for aws.neptune.db_subnet_group '{resource_id}'"
                    f"Describe will skip this aws.neptune.db_subnet_group and continue. "
                )
                continue
            tags = tags["ret"]
            resource_translated = hub.tool.aws.neptune.db_subnet_group.convert_raw_db_subnet_group_to_present(
                db_subnet_group, idem_resource_name=resource_id, tags=tags
            )
            result[resource_id] = {
                "aws.neptune.db_subnet_group.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }

    return result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Base class for implementing custom waiters for services that don't already have
prebuilt waiters. This class leverages botocore waiter code.
"""
import logging

import botocore.waiter

logger = logging.getLogger(__name__)


def waiter_wrapper(hub, **kwargs):
    class CustomWaiter:
        """
        Base class for a custom waiter that leverages botocore's waiter code. Waiters
        poll an operation, with a specified delay between each polling attempt, until
        either an accepted result is returned or the number of maximum attempts is reached.

        To use, implement a subclass that passes the specific operation, arguments,
        and acceptors to the superclass.

        For example, to implement a custom waiter for the transcription client that
        waits for both success and failure outcomes of the get_transcription_job function,
        create a class like the following:

            class TranscribeCompleteWaiter(CustomWaiter):
            def __init__(self, client):
                super().__init__(
                    'TranscribeComplete', 'GetTranscriptionJob',
                    'TranscriptionJob.TranscriptionJobStatus',
                    {'COMPLETED': WaitState.SUCCESS, 'FAILED': WaitState.FAILURE},
                    client)

            def wait(self, job_name):
                self._wait(TranscriptionJobName=job_name)

        """

        def __init__(self):
            """
            Subclasses should pass specific operations, arguments, and acceptors to
            their superclass.

            :param name: The name of the waiter. This can be any descriptive string.
            :param operation: The operation to wait for. This must match the casing of
                              the underlying operation model, which is typically in
                              CamelCase.
            :param argument: The dict keys used to access the result of the operation, in
                             dot notation. For example, 'Job.Status' will access
                             result['Job']['Status'].
            :param acceptors: The list of acceptors that indicate the wait is over. These
                              can indicate either success or failure. The acceptor values
                              are compared to the result of the operation after the
                              argument keys are applied.
            :param client: The Boto3 client.
            :param delay: The number of seconds to wait between each call to the operation.
            :param max_tries: The maximum number of tries before exiting.
            :param matcher: The kind of matcher to use.
            """
            self.name = kwargs.get("name")
            self.operation = kwargs.get("operation")
            self.argument = kwargs.get("argument")
            self.client = kwargs.get("client")
            self.acceptors = kwargs.get("acceptors")
            self.matcher = kwargs.get("matcher")
            self.delay = kwargs.get("delay")
            self.max_tries = kwargs.get("max_tries")
            self.waiter_model = botocore.waiter.WaiterModel(
                {
                    "version": 2,
                    "waiters": {
                        self.name: {
                            "delay": self.delay,
                            "operation": self.operation,
                            "maxAttempts": self.max_tries,
                            "acceptors": self.acceptors,
                        }
                    },
                }
            )
            self.waiter = botocore.waiter.create_waiter_with_client(
                self.name, self.waiter_model, self.client
            )

        def __call__(self, parsed, **kwargs):
            """
            Handles the after-call event by logging information about the operation and its
            result.

            :param parsed: The parsed response from polling the operation.
            :param kwargs: Not used, but expected by the caller.
            """
            status = parsed
            for arg in self.argument:
                for key in arg.split("."):
                    if key.endswith("[]") and key[:-2] in status:
                        status = status.get(key[:-2])[0]
                    elif key in status:
                        status = status.get(key)

                logger.info(
                    "Waiter %s called %s, got %s.", self.name, self.operation, status
                )

        def _wait(self, **kwargs):
            """
            Registers for the after-call event and starts the botocore wait loop.

            :param kwargs: Keyword arguments that are passed to the operation being polled.
            """
            event_name = f"after-call.{self.client.meta.service_model.service_name}"
            self.client.meta.events.register(event_name, self)
            self.waiter.wait(**kwargs)
            self.client.meta.events.unregister(event_name, self)

        def wait(self, **kwargs):
            self._wait(**kwargs)

    return CustomWaiter()



__contracts__ = ["soft_fail"]


async def lookup(hub, ctx, service: str, resource: str, filter: str = "*[]"):
    """
    Return the first match to the JMESPath filter on the named resource type.

    :param hub:
    :param ctx:
    :param service: The name of the service under the boto3 client to use
    :param resource: The name of the collection under the service to search
    :param filter: A JMESPath

    .. code-block:: bash

        $ idem exec aws.data.lookup service=ec2 resource=vpc filter="*[?CidrBlock==10.0.0.0/24]"
    """
    ret = dict(comment="", result=True, ret=None)
    resource = await hub.pop.loop.wrap(
        hub.tool.boto3.client.search,
        ctx,
        service_name=service,
        collection=resource,
        jmes_path=filter,
    )
    if not resource:
        ret["result"] = False
        ret["comment"] = f"Could not find '{service}.{resource}' with filter: {filter}"
    else:
        ret["comment"] = f"Found resource '{service}.{resource}'"
        ret["ret"] = resource
    return ret



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.neptune.copy_db_parameter_group
hub.exec.boto3.client.neptune.create_db_parameter_group
hub.exec.boto3.client.neptune.delete_db_parameter_group
hub.exec.boto3.client.neptune.describe_db_parameter_groups
hub.exec.boto3.client.neptune.modify_db_parameter_group
hub.exec.boto3.client.neptune.reset_db_parameter_group
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_parameter_group_family: str,
    parameters: List = None,
    resource_id: str = None,
    description: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new DB parameter group. A DB parameter group is initially created with the default parameters for the
    database engine used by the DB instance. To provide custom values for any of the parameters, you must modify the
    group after creating it using ModifyDBParameterGroup. Once you've created a DB parameter group, you need to
    associate it with your DB instance using ModifyDBInstance. When you associate a new DB parameter group with a
    running DB instance, you need to reboot the DB instance without failover for the new DB parameter group and
    associated settings to take effect.  After you create a DB parameter group, you should wait at least 5 minutes
    before creating your first DB instance that uses that DB parameter group as the default parameter group. This
    allows Amazon Neptune to fully complete the create action before the parameter group is used as the default for
    a new DB instance. This is especially important for parameters that are critical when creating the default
    database for a DB instance, such as the character set for the default database defined by the
    character_set_database parameter. You can use the Parameter Groups option of the Amazon Neptune console or the
    DescribeDBParameters command to verify that your DB parameter group has been created or modified.

    Args:
        name(Text): An idem name of the resource.
        db_parameter_group_family(Text): The DB parameter group family name. A DB parameter group can be associated with one and only one
            DB parameter group family, and can be applied only to a DB instance running a database engine
            and engine version compatible with that DB parameter group family.
        parameters(List, optional): An array of parameter names, values, and the apply method for the parameter update.
            At least one parameter name, value, and apply method must be supplied; subsequent arguments are optional. Defaults to None.
        resource_id(Text, optional): AWS DBParameterGroup Name. Defaults to None.
        description(Text): The description for the DB parameter group.
        tags(List, optional): The tags to be assigned to the new DB parameter group. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.neptune.db_parameter_group.present:
                - name: value
                - db_parameter_group_name: value
                - db_parameter_group_family: value
                - description: value
                - parameters:
                  - AllowedValues: 10-2147483647
                    ApplyMethod: pending-reboot
                    ApplyType: static
                    DataType: integer
                    Description: Graph query timeout (ms).
                    IsModifiable: true
                    MinimumEngineVersion: 1.0.1.0
                    ParameterName: neptune_query_timeout
                    ParameterValue: '120001'
                    Source: user
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.neptune.describe_db_parameter_groups(
            ctx, DBParameterGroupName=resource_id
        )
        if not before["result"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

    if before and before["ret"].get("DBParameterGroups"):
        resource_arn = before["ret"]["DBParameterGroups"][0].get("DBParameterGroupArn")
        old_tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not old_tags["result"]:
            result["result"] = False
            result["comment"] = old_tags["comment"]
            return result
        old_tags = old_tags["ret"]
        parameter_ret = await hub.exec.boto3.client.neptune.describe_db_parameters(
            ctx, DBParameterGroupName=resource_id
        )
        if not parameter_ret["result"]:
            result["result"] = False
            result["comment"] = parameter_ret["comment"]
            return result
        resource_translated = hub.tool.aws.neptune.db_parameter_group.convert_raw_db_parameter_group_to_present(
            before["ret"]["DBParameterGroups"][0],
            idem_resource_name=resource_id,
            tags=old_tags,
            parameters=parameter_ret["ret"].get("Parameters"),
        )
        result["old_state"] = resource_translated
        plan_state = copy.deepcopy(result["old_state"])

        old_parameters = result["old_state"].get("parameters", [])
        if parameters and not hub.tool.aws.state_comparison_utils.are_lists_identical(
            old_parameters, parameters
        ):
            resource_updated = True
            if ctx.get("test", False):
                plan_state["parameters"] = parameters
            else:
                ret = await hub.exec.boto3.client.neptune.modify_db_parameter_group(
                    ctx, DBParameterGroupName=name, Parameters=parameters
                )
                result["result"] = ret["result"]
                result["comment"] = result["comment"] + ret["comment"]
                if not result["result"]:
                    return result
        if (
            tags is not None
            and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                tags, old_tags
            )
        ):
            resource_updated = True
            update_ret = await hub.exec.aws.neptune.tag.update_tags(
                ctx=ctx,
                resource_arn=resource_arn,
                old_tags=old_tags,
                new_tags=tags,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                hub.log.debug(
                    f"Failed updating tags for aws.neptune.db_parameter_group '{name}'"
                )
                return result
            result["comment"] = result["comment"] + update_ret["comment"]
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["tags"] = update_ret["ret"].get("tags")

        if ctx.get("test", False):
            if resource_updated:
                result["new_state"] = plan_state
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.neptune.db_parameter_group", name=name
                )
            else:
                result["comment"] = (
                    f"aws.neptune.db_parameter_group '{name}' already exists",
                )
                result["new_state"] = copy.deepcopy(result["old_state"])
            return result
        if not resource_updated:
            result["comment"] = (
                f"aws.neptune.db_parameter_group '{name}' already exists",
            )
        else:
            result["comment"] = result["comment"] + (
                f"Updated aws.neptune.db_parameter_group '{name}'",
            )
    else:
        if ctx.get("test", False):
            # when a db_parameter_group is created aws will set some default parameters
            # this may change in the future so this code returns a fake parameter group
            default_parameters = [
                {
                    "AllowedValues": "fake",
                    "ApplyMethod": "pending-reboot",
                    "ApplyType": "static",
                    "DataType": "string",
                    "Description": "fake",
                    "IsModifiable": True,
                    "MinimumEngineVersion": "1.1.1.0",
                    "ParameterName": "fake_parameter",
                    "ParameterValue": "fake",
                    "Source": "fake",
                }
            ]
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "tags": tags,
                    "description": description,
                    "db_parameter_group_family": db_parameter_group_family,
                    "parameters": parameters if parameters else default_parameters,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.neptune.db_parameter_group", name=name
            )
            return result
        ret = await hub.exec.boto3.client.neptune.create_db_parameter_group(
            ctx,
            DBParameterGroupName=name,
            DBParameterGroupFamily=db_parameter_group_family,
            Description=description,
            Tags=tags,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["DBParameterGroup"]["DBParameterGroupName"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.neptune.db_parameter_group", name=name
        )
        if parameters:
            # have to modify it after creation to add the parameters
            # cant create it with the parameters
            ret = await hub.exec.boto3.client.neptune.modify_db_parameter_group(
                ctx, DBParameterGroupName=name, Parameters=parameters
            )
            result["result"] = ret["result"]
            result["comment"] = result["comment"] + ret["comment"]
            if not result["result"]:
                return result
    if (not before) or resource_updated:
        after = await hub.exec.boto3.client.neptune.describe_db_parameter_groups(
            ctx, DBParameterGroupName=resource_id
        )
        if not (after["result"] and after["ret"].get("DBParameterGroups")):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result
        resource_arn = (
            after["ret"].get("DBParameterGroups")[0].get("DBParameterGroupArn")
        )
        tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not tags["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + tags["comment"]
            return result
        tags = tags["ret"]
        parameter_ret = await hub.exec.boto3.client.neptune.describe_db_parameters(
            ctx, DBParameterGroupName=resource_id
        )
        if not parameter_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + parameter_ret["comment"]
            return result
        resource_translated = hub.tool.aws.neptune.db_parameter_group.convert_raw_db_parameter_group_to_present(
            after["ret"].get("DBParameterGroups")[0],
            idem_resource_name=name,
            parameters=parameter_ret["ret"].get("Parameters"),
            tags=tags,
        )
        result["new_state"] = resource_translated
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a specified DBParameterGroup. The DBParameterGroup to be deleted can't be associated with any DB
    instances.

    Args:
        name(Text): An idem name of the resource.
        resource_id(Text, Optional): AWS Neptune DB Parameter Group Name. Defaults to None
            Idem automatically considers this resource as absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.neptune.db_parameter_group.absent:
                - name: value
                - resource_id: value
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.neptune.db_parameter_group", name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.boto3.client.neptune.describe_db_parameter_groups(
        ctx, DBParameterGroupName=resource_id
    )
    db_parameter_groups = (
        before["ret"].get("DBParameterGroups") if before.get("ret") else None
    )

    if not (before["result"] and db_parameter_groups):
        # This condition means an error other than the resource not being found has happened
        if not ("DBParameterGroupNotFoundFault" in str(before["comment"])):
            result["comment"] = before["comment"]
            result["result"] = False
        return result

    resource_arn = db_parameter_groups[0].get("DBParameterGroupArn")
    tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
        ctx, resource_arn=resource_arn
    )
    if not tags["result"]:
        result["result"] = False
        result["comment"] = tags["comment"]
        return result
    tags = tags["ret"]
    parameter_ret = await hub.exec.boto3.client.neptune.describe_db_parameters(
        ctx, DBParameterGroupName=resource_id
    )
    if not parameter_ret["result"]:
        result["result"] = False
        result["comment"] = parameter_ret["comment"]
        return result
    resource_translated = hub.tool.aws.neptune.db_parameter_group.convert_raw_db_parameter_group_to_present(
        db_parameter_groups[0],
        idem_resource_name=resource_id,
        tags=tags,
        parameters=parameter_ret["ret"].get("Parameters"),
    )
    result["old_state"] = resource_translated

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.neptune.db_parameter_group", name=name
        )
    else:
        ret = await hub.exec.boto3.client.neptune.delete_db_parameter_group(
            ctx, DBParameterGroupName=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.neptune.db_parameter_group", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of DBParameterGroup descriptions.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.neptune.db_parameter_group
    """

    result = {}
    ret = await hub.exec.boto3.client.neptune.describe_db_parameter_groups(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe db_parameter_group {ret['comment']}")
        return {}

    for db_parameter_group in ret["ret"]["DBParameterGroups"]:
        resource_id = db_parameter_group.get("DBParameterGroupName")
        resource_arn = db_parameter_group.get("DBParameterGroupArn")
        tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not tags["result"]:
            # if something goes wrong fetching the tags (not if the tags are None that is a normal path)
            hub.log.warning(
                f"Failed listing tags for aws.neptune.db_parameter_group '{resource_id}'"
                f"Describe will skip this aws.neptune.db_parameter_group and continue. "
            )
            continue
        parameter_ret = await hub.exec.boto3.client.neptune.describe_db_parameters(
            ctx, DBParameterGroupName=resource_id
        )
        if not parameter_ret["result"]:
            hub.log.warning(
                f"Failed listing parameters for aws.neptune.db_parameter_group '{resource_id}'"
                f"Describe will skip this aws.neptune.db_parameter_group and continue. "
            )
            continue
        tags = tags["ret"]
        resource_translated = hub.tool.aws.neptune.db_parameter_group.convert_raw_db_parameter_group_to_present(
            db_parameter_group,
            idem_resource_name=resource_id,
            parameters=parameter_ret["ret"].get("Parameters"),
            tags=tags,
        )
        result[resource_id] = {
            "aws.neptune.db_parameter_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.elasticache.create_cache_subnet_group
hub.exec.boto3.client.elasticache.delete_cache_subnet_group
hub.exec.boto3.client.elasticache.describe_cache_subnet_groups
hub.exec.boto3.client.elasticache.modify_cache_subnet_group
"""
import copy
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
TREQ = {
    "present": {
        "require": [
            "aws.ec2.subnet.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    subnet_ids: List,
    cache_subnet_group_description: str,
    resource_id: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new cache subnet group. Use cache subnet group parameter only when you are creating a cluster in an Amazon Virtual
    Private Cloud (Amazon VPC).

    Args:
        name(Text): An Idem name of the resource.
        cache_subnet_group_description(Text): Description of the cache subnet group
        subnet_ids(list): A list of VPC subnet IDs for the cache subnet group.
        resource_id(Text, optional): AWS Elasticache subnet group name to identify the resource
        tags(List or Dict, optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                {tag-key: tag-value} The tags to assign to the cache_subnet_group. Defaults to None.

    Request Syntax:
        [cache_subnet_group_name]:
          aws.elasticache.cache_subnet_group.present:
          - name: 'string'
          - cache_subnet_group_description: 'string'
          - subnet_ids: 'List'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            cache_subnet_group_name:
              aws.elasticache.cache_subnet_group.present:
                - cache_subnet_group_description: "testing demo"
                - subnet_ids:
                    - subnet-239e4b44
                - tags:
                    - Key: Name
                      Value: test-elasticache
    """

    result = dict(comment=(), name=name, old_state=None, new_state=None, result=True)
    before = None
    resource_updated = False
    are_tags_updated = False
    resource_arn = str
    if resource_id:
        before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
            ctx, CacheSubnetGroupName=resource_id
        )
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    try:
        if before and before["result"] and before["ret"]["CacheSubnetGroups"]:
            convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
                ctx=ctx,
                raw_resource=before["ret"]["CacheSubnetGroups"][0],
                idem_resource_name=name,
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]
            plan_state = copy.deepcopy(result["old_state"])
            if cache_subnet_group_description != result["old_state"].get(
                "cache_subnet_group_description"
            ) or not hub.tool.aws.state_comparison_utils.are_lists_identical(
                subnet_ids, result["old_state"].get("subnet_ids")
            ):
                if not ctx.get("test", False):
                    update_ret = await hub.exec.boto3.client.elasticache.modify_cache_subnet_group(
                        ctx=ctx,
                        CacheSubnetGroupName=name,
                        CacheSubnetGroupDescription=cache_subnet_group_description,
                        SubnetIds=subnet_ids,
                    )
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = update_ret["result"]
                    resource_updated = resource_updated or bool(update_ret["ret"])
                else:
                    if cache_subnet_group_description is not None:
                        plan_state[
                            "cache_subnet_group_description"
                        ] = cache_subnet_group_description
                    if subnet_ids is not None:
                        plan_state["subnet_ids"] = subnet_ids
            resource_arn = before["ret"]["CacheSubnetGroups"][0]["ARN"]
            if tags is not None and tags != result["old_state"].get("tags"):
                # Update tags
                update_ret = await hub.exec.aws.elasticache.tag.update_tags(
                    ctx=ctx,
                    resource_arn=resource_arn,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]
                are_tags_updated = bool(update_ret["result"])
                resource_updated = resource_updated or are_tags_updated
                if ctx.get("test", False) and are_tags_updated:
                    plan_state["tags"] = update_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update tags for aws.elasticache.cache_subnet_group {name}",
                    )
        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "cache_subnet_group_description:": cache_subnet_group_description,
                        "subnet_ids": subnet_ids,
                        "tags": tags,
                    },
                )
                result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.elasticache.cache_subnet_group", name=name
                )
                return result

            ret = await hub.exec.boto3.client.elasticache.create_cache_subnet_group(
                ctx,
                CacheSubnetGroupName=name,
                CacheSubnetGroupDescription=cache_subnet_group_description,
                SubnetIds=subnet_ids,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            resource_id = name
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_arn = ret["ret"]["CacheSubnetGroup"]["ARN"]
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.elasticache.cache_subnet_group", name=name
            )
    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = (
                await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
                    ctx, CacheSubnetGroupName=resource_id
                )
            )
            if after and after.get("ret"):
                convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
                    ctx=ctx,
                    raw_resource=after["ret"]["CacheSubnetGroups"][0],
                    idem_resource_name=name,
                )
                result["result"] = convert_ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + convert_ret["comment"]
                result["new_state"] = convert_ret.get("ret")
            else:
                result["result"] = result["result"] and after["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + after["comment"]
                    return result
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a cache subnet group.  You cannot delete a default cache subnet group or one that is associated with any
    clusters.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): AWS Elasticache subnet group name to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.elasticache.cache_subnet_group.absent:
                - name: value
    """

    result = dict(comment=(), name=name, old_state=None, new_state=None, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.elasticache.cache_subnet_group", name=name
        )
        return result

    before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
        ctx, CacheSubnetGroupName=resource_id
    )
    if not before or not before["result"] or not before["ret"].get("CacheSubnetGroups"):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.elasticache.cache_subnet_group", name=name
        )
        return result
    else:
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
            ctx=ctx,
            raw_resource=before["ret"]["CacheSubnetGroups"][0],
            idem_resource_name=name,
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.elasticache.cache_subnet_group", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.elasticache.delete_cache_subnet_group(
                ctx, CacheSubnetGroupName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.elasticache.cache_subnet_group", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of cache subnet group descriptions. If a subnet group name is specified, the list contains only
    the description of that group. This is applicable only when you have ElastiCache in VPC setup. All ElastiCache
    clusters now launch in VPC by default.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.elasticache.cache_subnet_group
    """

    result = {}
    ret = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe cache_subnet_groups {ret['comment']}")
        return {}

    for resource in ret["ret"]["CacheSubnetGroups"]:
        resource_id = resource.get("CacheSubnetGroupName")
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
            ctx=ctx,
            raw_resource=resource,
            idem_resource_name=resource_id,
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe elasticache subnet group '{resource_id}' with error {convert_ret['comment']}"
            )
            continue
        resource_translated = convert_ret.get("ret")
        result[resource_id] = {
            "aws.elasticache.cache_subnet_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

"""
Util functions for AWS API Gateway v2 Domain Name resources.
"""


def convert_raw_domain_name_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 Domain Name resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(Dict[str, Any]): The AWS response to convert.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "ApiMappingSelectionExpression": "api_mapping_selection_expression",
            "DomainNameConfigurations": "domain_name_configurations",
            "MutualTlsAuthentication": "mutual_tls_authentication",
            "Tags": "tags",
        }
    )
    resource_translated = {
        "resource_id": raw_resource.get("DomainName"),
        "name": raw_resource.get("DomainName"),
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


def are_domain_name_configurations_identical(
    hub,
    new_configuration: List,
    old_configuration: List,
) -> bool:
    r"""
    Compares the new and old domain name configurations.

    Args:
        hub: required for functions in hub.
        new_configuration(List): The new domain name configuration parameters.
        old_configuration(List): The old domain name configuration parameters.

    Returns:
        bool: true if there are no differences between the new and old domain name configurations.
    """

    if (new_configuration is None or len(new_configuration) == 0) and (
        old_configuration is None or len(old_configuration) == 0
    ):
        return True
    if (
        new_configuration is None
        or len(new_configuration) == 0
        or old_configuration is None
        or len(old_configuration) == 0
    ):
        return False

    diff = [
        i
        for i in new_configuration + old_configuration
        if i not in new_configuration or i not in old_configuration
    ]

    return len(diff) == 0



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.efs.create_file_system
hub.exec.boto3.client.efs.delete_file_system
hub.exec.boto3.client.efs.describe_file_systems
hub.exec.boto3.client.efs.update_file_system
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    creation_token: str,
    resource_id: str = None,
    performance_mode: str = None,
    encrypted: bool = None,
    kms_key_id: str = None,
    throughput_mode: str = None,
    provisioned_throughput_in_mibps: float = None,
    availability_zone_name: str = None,
    backup: bool = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new, empty file system. The operation requires a creation token in the request that Amazon EFS uses to
    ensure idempotent creation (calling the operation with same creation token has no effect). If a file system does
    not currently exist that is owned by the caller's Amazon Web Services account with the specified creation token,
    this operation does the following:   Creates a new, empty file system. The file system will have an Amazon EFS
    assigned ID, and an initial lifecycle state creating.   Returns with the description of the created file system.
    Otherwise, this operation returns a FileSystemAlreadyExists error with the ID of the existing file system.  For
    basic use cases, you can use a randomly generated UUID for the creation token.   The idempotent operation allows
    you to retry a CreateFileSystem call without risk of creating an extra file system. This can happen when an
    initial call fails in a way that leaves it uncertain whether or not a file system was actually created. An
    example might be that a transport level timeout occurred or your connection was reset. As long as you use the
    same creation token, if the initial call had succeeded in creating a file system, the client can learn of its
    existence from the FileSystemAlreadyExists error. For more information, see Creating a file system in the Amazon
    EFS User Guide.  The CreateFileSystem call returns while the file system's lifecycle state is still creating.
    You can check the file system creation status by calling the DescribeFileSystems operation, which among other
    things returns the file system state.  This operation accepts an optional PerformanceMode parameter that you
    choose for your file system. We recommend generalPurpose performance mode for most file systems. File systems
    using the maxIO performance mode can scale to higher levels of aggregate throughput and operations per second
    with a tradeoff of slightly higher latencies for most file operations. The performance mode can't be changed
    after the file system has been created. For more information, see Amazon EFS performance modes. You can set the
    throughput mode for the file system using the ThroughputMode parameter. After the file system is fully created,
    Amazon EFS sets its lifecycle state to available, at which point you can create one or more mount targets for
    the file system in your VPC. For more information, see CreateMountTarget. You mount your Amazon EFS file system
    on an EC2 instances in your VPC by using the mount target. For more information, see Amazon EFS: How it Works.
    This operation requires permissions for the elasticfilesystem:CreateFileSystem action.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): EFS file system ID.
        creation_token(Text): A string of up to 64 ASCII characters. Amazon EFS uses this to ensure idempotent creation.
        performance_mode(Text, optional): The performance mode of the file system. We recommend generalPurpose performance mode for most
            file systems. File systems using the maxIO performance mode can scale to higher levels of
            aggregate throughput and operations per second with a tradeoff of slightly higher latencies for
            most file operations. The performance mode can't be changed after the file system has been
            created.  The maxIO mode is not supported on file systems using One Zone storage classes. Defaults to None.
        encrypted(bool, optional): A Boolean value that, if true, creates an encrypted file system. When creating an encrypted file
            system, you have the option of specifying an existing Key Management Service key (KMS key). If
            you don't specify a KMS key, then the default KMS key for Amazon EFS, /aws/elasticfilesystem, is
            used to protect the encrypted file system. Defaults to None.
        kms_key_id(Text, optional): The ID of the KMS key that you want to use to protect the encrypted file system. This parameter
            is only required if you want to use a non-default KMS key. If this parameter is not specified,
            the default KMS key for Amazon EFS is used. You can specify a KMS key ID using the following
            formats:   Key ID - A unique identifier of the key, for example
            1234abcd-12ab-34cd-56ef-1234567890ab.   ARN - An Amazon Resource Name (ARN) for the key, for
            example arn:aws:kms:us-west-2:111122223333:key/1234abcd-12ab-34cd-56ef-1234567890ab.   Key alias
            - A previously created display name for a key, for example alias/projectKey1.   Key alias ARN -
            An ARN for a key alias, for example arn:aws:kms:us-west-2:444455556666:alias/projectKey1.   If
            you use KmsKeyId, you must set the CreateFileSystemRequest$Encrypted parameter to true.  EFS
            accepts only symmetric KMS keys. You cannot use asymmetric KMS keys with Amazon EFS file
            systems. Defaults to None.
        throughput_mode(Text, optional): Specifies the throughput mode for the file system, either bursting or provisioned. If you set
            ThroughputMode to provisioned, you must also set a value for ProvisionedThroughputInMibps. After
            you create the file system, you can decrease your file system's throughput in Provisioned
            Throughput mode or change between the throughput modes, as long as it’s been more than 24 hours
            since the last decrease or throughput mode change. For more information, see Specifying
            throughput with provisioned mode in the Amazon EFS User Guide.  Default is bursting. Defaults to None.
        provisioned_throughput_in_mibps(float, optional): The throughput, measured in MiB/s, that you want to provision for a file system that you're
            creating. Valid values are 1-1024. Required if ThroughputMode is set to provisioned. The upper
            limit for throughput is 1024 MiB/s. To increase this limit, contact Amazon Web Services Support.
            For more information, see Amazon EFS quotas that you can increase in the Amazon EFS User Guide. Defaults to None.
        availability_zone_name(Text, optional): Used to create a file system that uses One Zone storage classes. It specifies the Amazon Web
            Services Availability Zone in which to create the file system. Use the format us-east-1a to
            specify the Availability Zone. For more information about One Zone storage classes, see Using
            EFS storage classes in the Amazon EFS User Guide.  One Zone storage classes are not available in
            all Availability Zones in Amazon Web Services Regions where Amazon EFS is available. Defaults to None.
        backup(bool, optional): Specifies whether automatic backups are enabled on the file system that you are creating. Set
            the value to true to enable automatic backups. If you are creating a file system that uses One
            Zone storage classes, automatic backups are enabled by default. For more information, see
            Automatic backups in the Amazon EFS User Guide. Default is false. However, if you specify an
            AvailabilityZoneName, the default is true.  Backup is not available in all Amazon Web Services
            Regions where Amazon EFS is available. Defaults to None.
        tags(List or Dict, optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                {tag-key: tag-value} The tags to assign to the file_system. Defaults to None.

    Request Syntax:
       [file-system-resource-id]:
            aws.efs.file_system.present:
            - name: 'string'
            - resource_id: 'string'
            - performance_mode: generalPurpose
            - encrypted: 'boolean'
            - kms_key_id: 'string'
            - creation_token: 'string'
            - throughput_mode: bursting/provisioned
            - provisioned_throughput_in_mibps: 'float'
            - availability_zone_name: 'string'
            - backup: 'boolean'
            - tags: List

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            fs-xxxxxxxxx:
              aws.efs.file_system.present:
              - name: efs-name
              - resource_id: fs-xxxxxxxxx
              - performance_mode: generalPurpose
              - encrypted: true
              - kms_key_id: arn:aws:kms:us-west-1:5xxxxxx:key/597f1c02-94dc-4c88-9088-f9aca5e39cf7
              - creation_token: xxx
              - throughput_mode: bursting
              - provisioned_throughput_in_mibps: 500.0
              - availability_zone_name: us-west-1b
              - backup: true
              - tags:
                - key: Name
                  value: efs-name

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Query for filesystem
    file_systems = None
    if resource_id:
        file_systems = await hub.exec.boto3.client.efs.describe_file_systems(
            ctx, FileSystemId=resource_id
        )
        if not file_systems["result"]:
            result["comment"] = file_systems["comment"]
            result["result"] = file_systems["result"]
            return result

    # Check for existing filesystem
    before = None
    if file_systems and file_systems["ret"]["FileSystems"]:
        before = file_systems["ret"]["FileSystems"][0]

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    backup_enabled = False
    if before:
        backup_result = await hub.exec.boto3.client.efs.describe_backup_policy(
            ctx,
            FileSystemId=resource_id,
        )
        if not backup_result[
            "result"
        ] and not hub.tool.aws.efs.file_system_utils.is_backup_policy_not_found(
            backup_result=backup_result
        ):
            hub.log.debug(
                f"Could not describe backup policy for file system {resource_id} : {backup_result['comment']}"
            )
            result["comment"] = backup_result["comment"]
            result["result"] = False
            return result

        backup_enabled = hub.tool.aws.efs.file_system_utils.is_backup_enabled(
            backup_result=backup_result
        )

    # Update current state
    current_state = (
        hub.tool.aws.efs.file_system_utils.convert_raw_file_system_to_present(
            file_system=before, backup_enabled=backup_enabled
        )
    )
    result["old_state"] = current_state

    # Handle No change behaviour
    desired_state = {
        "name": name,
        "resource_id": resource_id,
        "performance_mode": performance_mode,
        "encrypted": encrypted,
        "kms_key_id": kms_key_id,
        "creation_token": creation_token,
        "throughput_mode": throughput_mode,
        "provisioned_throughput_in_mibps": provisioned_throughput_in_mibps,
        "availability_zone_name": availability_zone_name,
        "backup": backup,
        "tags": tags,
    }

    if current_state:
        is_throughput_mode_updated = (
            hub.tool.aws.efs.file_system_utils.is_throughput_mode_updated(
                current_state=current_state, desired_state=desired_state
            )
        )

        is_tags_updated = (current_state.get("tags")) != tags

        is_backup_policy_updated = current_state.get("backup") != desired_state.get(
            "backup"
        )

    is_change_detected = (
        before is None
        or is_throughput_mode_updated
        or is_backup_policy_updated
        or is_tags_updated
    )

    if not is_change_detected:
        result["comment"] = (f"aws.efs.file_system '{name}' already exists",)
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.efs.file_system", name=name
            )
            if before
            else hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.efs.file_system", name=name
            )
        )
        return result

    # Handle actual resource create or update
    if before:
        # Update ThroughputMode and ProvisionedThroughputInMibps
        if is_throughput_mode_updated:
            ret = await hub.exec.boto3.client.efs.update_file_system(
                ctx,
                FileSystemId=resource_id,
                ThroughputMode=throughput_mode,
                ProvisionedThroughputInMibps=provisioned_throughput_in_mibps
                if throughput_mode == "provisioned"
                else None,
            )
            result["result"] = ret["result"]
            result["comment"] = (
                f"Updated ThroughputMode for aws.efs.file_system '{name}'",
            )
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

        # Update backup policy
        if is_backup_policy_updated:
            ret = await hub.exec.boto3.client.efs.put_backup_policy(
                ctx,
                FileSystemId=resource_id,
                BackupPolicy={"Status": "ENABLED" if backup else "DISABLED"},
            )
            result["result"] = result["result"] and ret["result"]
            result["comment"] = result["comment"] + (
                f"Updated Backup policy for aws.efs.file_system '{name}'",
            )
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

        # Update tags
        if is_tags_updated:
            update_tags = await hub.exec.aws.efs.tag.update_tags(
                ctx,
                resource_id=resource_id,
                old_tags=current_state["tags"],
                new_tags=desired_state["tags"],
            )
            result["comment"] = result["comment"] + update_tags["comment"]
            result["result"] = result["result"] and update_tags["result"]
            if not result["result"]:
                return result
    else:
        ret = await hub.exec.boto3.client.efs.create_file_system(
            ctx,
            **{
                "CreationToken": creation_token,
                "PerformanceMode": performance_mode,
                "Encrypted": encrypted,
                "KmsKeyId": kms_key_id,
                "ThroughputMode": throughput_mode,
                "ProvisionedThroughputInMibps": provisioned_throughput_in_mibps
                if throughput_mode == "provisioned"
                else None,
                "AvailabilityZoneName": availability_zone_name,
                "Backup": backup,
                "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.efs.file_system", name=name
        )
        resource_id = ret["ret"]["FileSystemId"]

    # Fetch the updated resource and update new_state
    file_systems = await hub.exec.boto3.client.efs.describe_file_systems(
        ctx, FileSystemId=resource_id
    )
    if not file_systems["result"]:
        result["comment"] = result["comment"] + file_systems["comment"]
        result["result"] = file_systems["result"]
        return result

    after = file_systems["ret"]["FileSystems"][0]
    backup_result = await hub.exec.boto3.client.efs.describe_backup_policy(
        ctx,
        FileSystemId=resource_id,
    )
    if not backup_result[
        "result"
    ] and not hub.tool.aws.efs.file_system_utils.is_backup_policy_not_found(
        backup_result=backup_result
    ):
        hub.log.debug(
            f"Could not describe backup policy for file system {resource_id} : {backup_result['comment']}"
        )
        result["comment"] = result["comment"] + backup_result["comment"]
        result["result"] = False
        return result

    backup_enabled = hub.tool.aws.efs.file_system_utils.is_backup_enabled(backup_result)
    result[
        "new_state"
    ] = hub.tool.aws.efs.file_system_utils.convert_raw_file_system_to_present(
        file_system=after, backup_enabled=backup_enabled
    )
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a file system, permanently severing access to its contents. Upon return, the file system no longer
    exists and you can't access any contents of the deleted file system.  You can't delete a file system that is in
    use. That is, if the file system has any mount targets, you must first delete them. For more information, see
    DescribeMountTargets and DeleteMountTarget.   The DeleteFileSystem call returns while the file system state is
    still deleting. You can check the file system deletion status by calling the DescribeFileSystems operation,
    which returns a list of file systems in your account. If you pass file system ID or creation token for the
    deleted file system, the DescribeFileSystems returns a 404 FileSystemNotFound error.  This operation requires
    permissions for the elasticfilesystem:DeleteFileSystem action.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): EFS file system ID.

    Request Syntax:
       [file-system-resource-id]:
            aws.efs.file_system.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            fs-xxxxxxxxx:
              aws.efs.file_system.absent:
              - name: efs-name
              - resource_id: fs-xxxxxxxxx
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    file_systems = None
    if resource_id:
        file_systems = await hub.exec.boto3.client.efs.describe_file_systems(
            ctx, FileSystemId=resource_id
        )

    # Check for existing filesystem
    before = None
    if file_systems and file_systems["result"] and file_systems["ret"]["FileSystems"]:
        before = file_systems["ret"]["FileSystems"][0]

    if before:
        backup_result = await hub.exec.boto3.client.efs.describe_backup_policy(
            ctx,
            FileSystemId=resource_id,
        )
        if not backup_result[
            "result"
        ] and not hub.tool.aws.efs.file_system_utils.is_backup_policy_not_found(
            backup_result=backup_result
        ):
            hub.log.debug(
                f"Could not describe backup policy for file system {resource_id} : {backup_result['comment']}"
            )
            result["comment"] = result["comment"] + backup_result["comment"]
            result["result"] = False
            return result
        backup_enabled = hub.tool.aws.efs.file_system_utils.is_backup_enabled(
            backup_result
        )
        result[
            "old_state"
        ] = hub.tool.aws.efs.file_system_utils.convert_raw_file_system_to_present(
            file_system=before, backup_enabled=backup_enabled
        )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.efs.file_system", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.efs.file_system", name=name
        )
    else:
        ret = await hub.exec.boto3.client.efs.delete_file_system(
            ctx, FileSystemId=resource_id
        )
        result["result"] = ret["result"]
        result["comment"] = (
            hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.efs.file_system", name=name
            )
            if result["result"]
            else (ret["comment"],)
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns the description of a specific Amazon EFS file system if either the file system CreationToken or the
    FileSystemId is provided. Otherwise, it returns descriptions of all file systems owned by the caller's Amazon
    Web Services account in the Amazon Web Services Region of the endpoint that you're calling. When retrieving all
    file system descriptions, you can optionally specify the MaxItems parameter to limit the number of descriptions
    in a response. Currently, this number is automatically set to 10. If more file system descriptions remain,
    Amazon EFS returns a NextMarker, an opaque token, in the response. In this case, you should send a subsequent
    request with the Marker request parameter set to the value of NextMarker.  To retrieve a list of your file
    system descriptions, this operation is used in an iterative process, where DescribeFileSystems is called first
    without the Marker and then the operation continues to call it with the Marker parameter set to the value of the
    NextMarker from the previous response until the response has no NextMarker.   The order of file systems returned
    in the response of one DescribeFileSystems call and the order of file systems returned across the responses of a
    multi-call iteration is unspecified.   This operation requires permissions for the
    elasticfilesystem:DescribeFileSystems action.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.efs.file_system
    """

    result = {}
    ret = await hub.exec.boto3.client.efs.describe_file_systems(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe file_system {ret['comment']}")
        return {}

    for file_system in ret["ret"]["FileSystems"]:
        resource_id = file_system["FileSystemId"]

        backup_result = await hub.exec.boto3.client.efs.describe_backup_policy(
            ctx,
            FileSystemId=resource_id,
        )
        if not backup_result[
            "result"
        ] and not hub.tool.aws.efs.file_system_utils.is_backup_policy_not_found(
            backup_result=backup_result
        ):
            hub.log.debug(
                f"Could not describe backup policy for file system {resource_id} : {backup_result['comment']}"
            )
            continue
        backup_enabled = hub.tool.aws.efs.file_system_utils.is_backup_enabled(
            backup_result
        )

        resource = (
            hub.tool.aws.efs.file_system_utils.convert_raw_file_system_to_present(
                file_system=file_system, backup_enabled=backup_enabled
            )
        )
        result[resource_id] = {
            "aws.efs.file_system.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    url: str,
    thumbprint_list: List,
    resource_id: str = None,
    client_id_list: List = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    Creates an IAM entity to describe an identity provider (IdP) that supports OpenID Connect (OIDC).

    The OIDC provider that you create with this operation can be used as a principal in a role's trust policy.
    Such a policy establishes a trust relationship between Amazon Web Services and the OIDC provider.

    If you are using an OIDC identity provider from Google, Facebook, or Amazon Cognito, you don't need to create a
    separate IAM identity provider. These OIDC identity providers are already built-in to Amazon Web Services and are
    available for your use. Instead, you can move directly to creating new roles using your identity provider. To learn
    more, see Creating a role for web identity or OpenID connect federation in the IAM User Guide.

    When you create the IAM OIDC provider, you specify the following:
        * The URL of the OIDC identity provider (IdP) to trust
        * A list of client IDs (also known as audiences) that identify the application or applications allowed to
          authenticate using the OIDC provider.
        * A list of thumbprints of one or more server certificates that the IdP uses

    You get all of this information from the OIDC IdP you want to use to access Amazon Web Services.

    Args:
        name(Text): The idem name for IAM OIDC provider.

        url(Text): The URL of the identity provider. The URL must begin with https:// and should correspond to the iss
            claim in the provider's OpenID Connect ID tokens. Per the OIDC standard, path components are allowed but
            query parameters are not. Typically the URL consists of only a hostname, like https://server.example.org or
            https://example.com . The URL should not contain a port number.

            You cannot register the same provider multiple times in a single Amazon Web Services account. If you try to
            submit a URL that has already been used for an OpenID Connect provider in the Amazon Web Services account,
            you will get an error.

        thumbprint_list(List): A list of server certificate thumbprints for the OpenID Connect (OIDC) identity
            provider's server certificates. Typically this list includes only one entry. However, IAM lets you have up
            to five thumbprints for an OIDC provider. This lets you maintain multiple thumbprints if the identity
            provider is rotating certificates.

            The server certificate thumbprint is the hex-encoded SHA-1 hash value of the X.509 certificate used by the
            domain where the OpenID Connect provider makes its keys available. It is always a 40-character string.

            You must provide at least one thumbprint when creating an IAM OIDC provider. For example, assume that the
            OIDC provider is server.example.com and the provider stores its keys at
            https://keys.server.example.com/openid-connect. In that case, the thumbprint string would be the hex-encoded
             SHA-1 hash value of the certificate used by https://keys.server.example.com.

            For more information about obtaining the OIDC provider thumbprint, see Obtaining the thumbprint for an
            OpenID Connect provider in the IAM User Guide .

            (string) --
                Contains a thumbprint for an identity provider's server certificate.

                The identity provider's server certificate thumbprint is the hex-encoded SHA-1 hash value of the
                self-signed X.509 certificate. This thumbprint is used by the domain where the OpenID Connect provider
                makes its keys available. The thumbprint is always a 40-character string.

        resource_id(Text, optional): The Amazon Resource Name (ARN) of the IAM OpenID Connect provider resource object.

        client_id_list(List, optional): Provides a list of client IDs, also known as audiences. When a mobile or web
            app registers with an OpenID Connect provider, they establish a value that identifies the application. This
            is the value that's sent as the client_id parameter on OAuth requests.

            You can register multiple client IDs with the same provider. For example, you might have multiple
            applications that use the same OIDC provider. You cannot register more than 100 client IDs with a single
            IAM OIDC provider.

            There is no defined format for a client ID. The CreateOpenIDConnectProviderRequest operation accepts
            client IDs up to 255 characters long.

        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the IAM OpenID Connect (OIDC) provider.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key): The key name that can be used to look up or retrieve the associated value. For example,
                Department or Cost Center are common choices.
            * (Value): The value associated with this tag. For example, tags with a key name of Department could have
                values such as Human Resources, Accounting, and Support. Tags with a key name of Cost Center
                might have values that consist of the number associated with the different cost centers in your
                company. Typically, many resources have tags with the same key name but with different values.
                Amazon Web Services always interprets the tag Value as a single string. If you need to store an
                array, you can store comma-separated values in the string. However, you must interpret the value
                in your code.
    Request Syntax:
        [oidc-resource-name]:
          aws.iam.open_id_connect_provider.present:
          - name: 'string'
          - resource_id: 'string'
          - url: 'string'
          - client_id_list:
            - 'string'
          - thumbprint_list:
            - 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            oidcprovider-1:
              aws.iam.open_id_connect_provider.present:
              - name: oidcprovider-1
              - url: https://abc.example.com
              - client_id_list:
                - acd.amazonaws.com
                - xyz.com
              - thumbprint_list:
                - 9e99a48a9960b14926bb7f3b02e22da2b0ab7402
              - tags:
                - Key: Name
                  Value: oidc-name
                - Key: env
                  Value: dev

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        response = await hub.exec.boto3.client.iam.get_open_id_connect_provider(
            ctx, OpenIDConnectProviderArn=resource_id
        )
        if not response["result"]:
            result["comment"] = response["comment"]
            result["result"] = response["result"]
            return result
        before = response["ret"]

    if before:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.iam.conversion_utils.convert_raw_oidc_provider_to_present(
                raw_resource=before, idem_resource_name=name, resource_id=resource_id
            )
            plan_state = copy.deepcopy(result["old_state"])

            # update clientIDs
            if client_id_list is not None:
                update_ret = (
                    await hub.exec.aws.iam.open_id_connect_provider.update_client_ids(
                        ctx=ctx,
                        resource_id=resource_id,
                        old_client_ids=result["old_state"].get("client_id_list", []),
                        new_client_ids=client_id_list,
                    )
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["client_id_list"] = update_ret["ret"].get(
                        "client_id_list"
                    )
                if not result["result"]:
                    return result

            # update thumbprints
            if (
                thumbprint_list is not None
                and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    result["old_state"].get("thumbprint_list"), thumbprint_list
                )
            ):
                update_ret = (
                    await hub.exec.aws.iam.open_id_connect_provider.update_thumbprints(
                        ctx=ctx,
                        resource_id=resource_id,
                        old_thumbprints=result["old_state"].get("thumbprint_list"),
                        new_thumbprints=thumbprint_list,
                    )
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["thumbprint_list"] = update_ret["ret"].get(
                        "thumbprint_list"
                    )
                if not result["result"]:
                    return result

            # update tags
            if tags is not None and tags != result["old_state"].get("tags"):
                update_ret = (
                    await hub.exec.aws.iam.open_id_connect_provider.update_tags(
                        ctx=ctx,
                        resource_id=resource_id,
                        old_tags=result["old_state"].get("tags"),
                        new_tags=tags,
                    )
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["result"])
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["tags"] = update_ret["ret"]
                if not result["result"]:
                    return result
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "url": url,
                    "client_id_list": client_id_list,
                    "thumbprint_list": thumbprint_list,
                    "tags": tags,
                },
            )
            result["comment"] = (
                f"Would create aws.iam.open_id_connect_provider '{name}'",
            )
            return result

        try:
            ret = await hub.exec.boto3.client.iam.create_open_id_connect_provider(
                ctx,
                Url=url,
                ClientIDList=client_id_list,
                ThumbprintList=thumbprint_list,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = (f"Created '{name}'",)
            resource_id = ret["ret"]["OpenIDConnectProviderArn"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.iam.get_open_id_connect_provider(
                ctx, OpenIDConnectProviderArn=resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.iam.conversion_utils.convert_raw_oidc_provider_to_present(
                raw_resource=after["ret"],
                idem_resource_name=name,
                resource_id=resource_id,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """
    Deletes an OpenID Connect identity provider (IdP) resource object in IAM.
    Deleting an IAM OIDC provider resource does not update any roles that reference the provider as a principal in their
    trust policies. Any attempt to assume a role that references a deleted provider fails.

    Args:
        name(Text): The idem name for IAM OIDC provider.
        resource_id(Text): The Amazon Resource Name (ARN) of the IAM OpenID Connect provider resource object to delete.

    Request Syntax:
        [oidc-resource-id]:
          aws.iam.open_id_connect_provider.absent:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            oidc-provider-1:
              aws.iam.open_id_connect_provider.absent:
                - name: oidc-provider-1
                - resource_id: arn:aws:iam::565284315989:oidc-provider/test.example.com

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.open_id_connect_provider", name=name
        )
        return result
    before = await hub.exec.boto3.client.iam.get_open_id_connect_provider(
        ctx, OpenIDConnectProviderArn=resource_id
    )

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.open_id_connect_provider", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_oidc_provider_to_present(
            raw_resource=before["ret"], idem_resource_name=name, resource_id=resource_id
        )
        result["comment"] = (f"Would delete aws.iam.open_id_connect_provider '{name}'",)
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_oidc_provider_to_present(
            raw_resource=before["ret"], idem_resource_name=name, resource_id=resource_id
        )
        try:
            ret = await hub.exec.boto3.client.iam.delete_open_id_connect_provider(
                ctx, OpenIDConnectProviderArn=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets information about the AWS IAM OpenID Connect provider

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.open_id_connect_provider

    """
    result = {}
    ret = await hub.exec.boto3.client.iam.list_open_id_connect_providers(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe open id connect providers {ret['comment']}")
        return {}

    for provider in ret["ret"]["OpenIDConnectProviderList"]:
        resource_id = provider.get("Arn")
        connect_provider_resp = (
            await hub.exec.boto3.client.iam.get_open_id_connect_provider(
                ctx, OpenIDConnectProviderArn=resource_id
            )
        )
        if not connect_provider_resp["result"]:
            hub.log.warning(
                f"Failed on fetching open_id_connect_provider with arn {resource_id} "
                f"with error {connect_provider_resp['comment']}. Describe will skip this open_id_connect_provider and continue."
            )
            continue

        resource_translated = (
            hub.tool.aws.iam.conversion_utils.convert_raw_oidc_provider_to_present(
                raw_resource=connect_provider_resp["ret"],
                idem_resource_name=resource_id,
                resource_id=resource_id,
            )
        )

        result[resource_id] = {
            "aws.iam.open_id_connect_provider.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS API Gateway v2 Stage resources.
"""


def convert_raw_stage_to_present(
    hub, api_id: str, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 Stage resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        api_id(string): The API resource identifier in Amazon Web Services.
        raw_resource(Dict[str, Any]): The AWS response to convert.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "AccessLogSettings": "access_log_settings",
            "ApiGatewayManaged": "api_gateway_managed",
            "AutoDeploy": "auto_deploy",
            "ClientCertificateId": "client_certificate_id",
            "DefaultRouteSettings": "default_route_settings",
            "DeploymentId": "deployment_id",
            "Description": "description",
            "LastDeploymentStatusMessage": "last_deployment_status_message",
            "RouteSettings": "route_settings",
            "StageVariables": "stage_variables",
            "Tags": "tags",
        }
    )
    resource_translated = {
        "resource_id": raw_resource.get("StageName"),
        "name": raw_resource.get("StageName"),
        "api_id": api_id,
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.sns.create_topic
hub.exec.boto3.client.sns.delete_topic
hub.exec.boto3.client.sns.list_topics
resource = hub.tool.boto3.resource.create(ctx, "sns", "Topic", name)
hub.tool.boto3.resource.exec(resource, add_permission, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, confirm_subscription, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, publish, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, remove_permission, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, set_attributes, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, subscribe, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    attributes: Dict = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a topic to which notifications can be published. Users can create at most 100,000 standard topics (at
    most 1,000 FIFO topics). For more information, see Creating an Amazon SNS topic in the Amazon SNS Developer
    Guide. This action is idempotent, so if the requester already owns a topic with the specified name, that topic's
    ARN is returned without creating a new topic.

    Args:
        hub:
        ctx:
        name(Text): The idem name for the resource
        resource_id(Text, optional): The AWS resource identifier, here it is resource arn
        attributes(Dict, optional): Attributes of topic
        tags(List or Dict, optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                {tag-key: tag-value}

    Request Syntax:
        [topic-name]:
          aws.sns.topic.present:
          - resource_id: 'string'
          - attributes: 'Dict'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            new-topic:
              aws.sns.topic.present:
              - attributes:
                  DeliveryPolicy: '{"http":{"defaultHealthyRetryPolicy":{"minDelayTarget":10,"maxDelayTarget":30,"numRetries":10,"numMaxDelayRetries":7,"numNoDelayRetries":0,"numMinDelayRetries":3,"backoffFunction":"linear"}}}'
              - tags:
                 - Key: Name
                 - Value: new-topic
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None
    # Standardise JSON string format
    if attributes:
        for key, value in attributes.items():
            attributes[key] = hub.tool.aws.sns.conversion_utils.standardise_json(value)
    if resource_id:
        resource = hub.tool.boto3.resource.create(ctx, "sns", "Topic", arn=resource_id)
        before = await hub.tool.boto3.resource.describe(resource)

    if before:
        try:
            ret_tags = await hub.exec.boto3.client.sns.list_tags_for_resource(
                ctx, ResourceArn=resource_id
            )
            if not ret_tags["result"]:
                result["comment"] = ret_tags["comment"]
                result["result"] = False
                return result

            result[
                "old_state"
            ] = hub.tool.aws.sns.conversion_utils.convert_raw_topic_to_present(
                raw_resource=before, raw_resource_tags=ret_tags, idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])
            old_attributes = plan_state.get("attributes")
            old_tags = plan_state.get("tags")
            if tags is not None:
                # Check and update tags
                update_tags_ret = await hub.exec.aws.sns.tag.update_tags(
                    ctx=ctx,
                    resource_arn=resource_id,
                    old_tags=old_tags,
                    new_tags=tags,
                )
                if not update_tags_ret["result"]:
                    result["comment"] = update_tags_ret["comment"]
                    result["result"] = False
                    return result
                resource_updated = bool(update_tags_ret["ret"])
                result["comment"] = update_tags_ret["comment"]
                if ctx.get("test", False) and update_tags_ret["ret"]:
                    plan_state["tags"] = update_tags_ret["ret"]

            if attributes is not None:
                # Check and update Attributes
                update_attributes_ret = (
                    await hub.exec.aws.sns.attributes.update_topic_attributes(
                        ctx=ctx,
                        resource_arn=resource_id,
                        old_attributes=old_attributes,
                        new_attributes=attributes,
                    )
                )
                if not update_attributes_ret["result"]:
                    result["comment"] = update_attributes_ret["comment"]
                    result["result"] = False
                    return result
                resource_updated = resource_updated or bool(
                    update_attributes_ret["ret"]
                )
                result["comment"] = result["comment"] + update_attributes_ret["comment"]
                if ctx.get("test", False) and update_attributes_ret["ret"]:
                    plan_state["attributes"] = update_attributes_ret["ret"].get(
                        "updated_attributes"
                    )

            else:
                result["comment"] = (f"aws.sns.topic '{name}' already exists",)

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    else:
        if tags is not None:
            if isinstance(tags, Dict):
                tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                tags_dict = tags
            else:
                tags_list = tags
                tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "attributes": attributes,
                    "tags": tags_dict,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.sns.topic", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.sns.create_topic(
                ctx, Name=name, Attributes=attributes, Tags=tags_list
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = ret["ret"].get("TopicArn")
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                "aws.sns.topic", name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            resource = hub.tool.boto3.resource.create(
                ctx, "sns", "Topic", arn=resource_id
            )
            after = await hub.tool.boto3.resource.describe(resource)
            ret_tags = await hub.exec.boto3.client.sns.list_tags_for_resource(
                ctx, ResourceArn=resource_id
            )
            if not ret_tags["result"]:
                result["comment"] = ret_tags["comment"]
                result["result"] = False
                return result
            result[
                "new_state"
            ] = hub.tool.aws.sns.conversion_utils.convert_raw_topic_to_present(
                raw_resource=after, raw_resource_tags=ret_tags, idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a topic and all its subscriptions. Deleting a topic might prevent some messages previously sent to the
    topic from being delivered to subscribers. This action is idempotent, so deleting a topic that does not exist
    does not result in an error.

    Args:
        name(Text): The name of the state.
        resource_id(Text): The AWS resource identifier, here it is resource arn

    Request Syntax:
        [topic-name]:
          aws.sns.topic.absent:
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test-topic:
              aws.sns.topic.absent:
                - resource_id: arn:aws:sns:eu-west-3:537227425989:test-topic
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "sns", "Topic", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            "aws.sns.topic", name
        )

    else:
        ret_tags = await hub.exec.boto3.client.sns.list_tags_for_resource(
            ctx, ResourceArn=resource_id
        )
        if not ret_tags["result"]:
            result["comment"] = ret_tags["comment"]
            result["result"] = False
            return result
        result[
            "old_state"
        ] = hub.tool.aws.sns.conversion_utils.convert_raw_topic_to_present(
            raw_resource=before, raw_resource_tags=ret_tags, idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                "aws.sns.topic", name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.sns.delete_topic(
                ctx, TopicArn=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                "aws.sns.topic", name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of the requester's topics. Each call returns a limited list of topics, up to 100. If there are
    more topics, a NextToken is also returned. Use the NextToken parameter in a new ListTopics call to get further
    results. This action is throttled at 30 transactions per second (TPS).


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.sns.topic
    """

    result = {}
    ret = await hub.exec.boto3.client.sns.list_topics(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe topic {ret['comment']}")
        return {}

    for topic in ret["ret"]["Topics"]:
        # Including fields to match the 'present' function parameters
        resource_id = topic["TopicArn"]
        idem_resource_name = resource_id.split(":")[-1]

        resource = hub.tool.boto3.resource.create(ctx, "sns", "Topic", arn=resource_id)
        raw_resource = await hub.tool.boto3.resource.describe(resource)
        ret_tags = await hub.exec.boto3.client.sns.list_tags_for_resource(
            ctx, ResourceArn=resource_id
        )
        if not ret_tags["result"]:
            hub.log.debug(f"Could not describe topic tag {ret_tags['comment']}")
            return {}

        resource_translated = (
            hub.tool.aws.sns.conversion_utils.convert_raw_topic_to_present(
                raw_resource=raw_resource,
                raw_resource_tags=ret_tags,
                idem_resource_name=idem_resource_name,
            )
        )
        result[idem_resource_name] = {
            "aws.sns.topic.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.upload_ssh_public_key
hub.exec.boto3.client.iam.delete_ssh_public_key
hub.exec.boto3.client.iam.list_ssh_public_keys
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    user_name: str,
    ssh_public_key_body: str,
    resource_id: str = None,
    status: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Uploads an SSH public key and associates it with the specified IAM user.

    The SSH public key uploaded by this operation can be used only for authenticating the associated IAM user to an
    CodeCommit repository.

    Args:
        name(Text): A name of the ssh public key to be added/modified.
        user_name(Text): The name (friendly name, not ARN) of the IAM user to associate the SSH public key with.
                         This parameter allows (through its regex pattern ) a string of characters consisting of upper
                         and lowercase alphanumeric characters with no spaces. You can also include any of the following characters: _+=,.@-
        ssh_public_key_body(Text): The SSH public key. The public key must be encoded in ssh-rsa format or PEM format.
                          The minimum bit-length of the public key is 2048 bits. For example, you can generate
                          a 2048-bit key, and the resulting PEM file is 1679 bytes long.
        resource_id(Text, Optional): The Id of the ssh public key to be added/modified.
        status (Text, Optional): The status of the SSH public key. The value can be 'Acitve' or 'Inactive'. Defaults 'Active'.

    Request Syntax:
        [iam-user-ssh-key-name]:
          aws.iam.user_ssh_key.present:
          - user_name: 'string'
          - ssh_public_key_body: 'string'
          - resource_id: 'string'
          - status: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls
        APKASYLXZO1EQCR0ZAL7:
          aws.iam.user_ssh_key.present:
          - user_name:  my-iam-user-1
          - ssh_public_key_body: ssh-rsa AAAB3adeaadasdamasdasdasdsadsdasdadasda8ad800q3q3=
          - resource_id: my-iam-user-id
          - status: Active

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    is_ssh_key_associated = False
    before_ret = None
    resource_updated = False
    if resource_id:
        before_ret = (
            await hub.exec.aws.iam.user_ssh_key.get_ssh_key_if_attached_to_user(
                ctx, user_name=user_name, ssh_public_key_id=resource_id
            )
        )
    if before_ret and before_ret["result"]:
        # ssh public key already exists for user. We can only modify status (Active | Inactive)
        # Check if status is modified and update if modified.
        ret_ssh_public_key_body = await hub.exec.boto3.client.iam.get_ssh_public_key(
            ctx=ctx,
            UserName=user_name,
            SSHPublicKeyId=resource_id,
            Encoding="SSH",
        )
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_ssh_key_to_present(
            raw_resource=before_ret["ret"],
            idem_resource_name=resource_id,
            ret_ssh_public_key_body=ret_ssh_public_key_body,
        )
        plan_state = copy.deepcopy(result["old_state"])
        if status is not None and result["old_state"].get("status") != status:
            if not ctx.get("test", False):
                update_ret = await hub.exec.boto3.client.iam.update_ssh_public_key(
                    ctx, UserName=user_name, SSHPublicKeyId=resource_id, Status=status
                )
                result["result"] = update_ret["result"]
                if not result["result"]:
                    result["comment"] = update_ret["comment"]
                    return result
                result["comment"] = (
                    f"Updated aws.iam.user_ssh_key '{name}' status to '{status}'",
                )
                resource_updated = resource_updated or bool(update_ret["ret"])
            else:
                plan_state["status"] = status
                result["comment"] = (
                    f"Would update ssh public key '{name}' of user '{user_name}' status to {status}",
                )
            is_ssh_key_associated = True

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "user_name": user_name,
                    "ssh_public_key_body": ssh_public_key_body,
                    "status": status,
                },
            )
            result["comment"] = (
                f"Would upload the ssh public key and attach it to iam user '{user_name}'",
            )
            return result
        ret = await hub.exec.boto3.client.iam.upload_ssh_public_key(
            ctx,
            UserName=user_name,
            SSHPublicKeyBody=ssh_public_key_body,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        name = ret["ret"].get("SSHPublicKey").get("SSHPublicKeyId")
        result["comment"] = (
            f"Uploaded the ssh public key and attached it to iam user '{user_name}'",
        )
        is_ssh_key_associated = True
    if is_ssh_key_associated:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before_ret) or resource_updated:
            after_ret = (
                await hub.exec.aws.iam.user_ssh_key.get_ssh_key_if_attached_to_user(
                    ctx, user_name=user_name, ssh_public_key_id=name
                )
            )
            if after_ret and after_ret["result"]:
                created_resource_id = after_ret["ret"].get("SSHPublicKeyId")
                ret_ssh_public_key_body = (
                    await hub.exec.boto3.client.iam.get_ssh_public_key(
                        ctx=ctx,
                        UserName=user_name,
                        SSHPublicKeyId=created_resource_id,
                        Encoding="SSH",
                    )
                )
                result[
                    "new_state"
                ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_ssh_key_to_present(
                    raw_resource=after_ret["ret"],
                    idem_resource_name=created_resource_id,
                    ret_ssh_public_key_body=ret_ssh_public_key_body,
                )
            else:
                result["comment"] = after_ret["comment"]
                result["result"] = False
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    user_name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified SSH public key.
    The SSH public key deleted by this operation is used only for authenticating
    the associated IAM user to an CodeCommit repository.

    Args:
        hub:
        ctx:
        name(Text): The name of the ssh public key to be deleted.
        user_name(Text): The name (friendly name, not ARN) of the IAM user to delete the ssh public key from.
        name(Text): The ID of the ssh public key to be deleted.
        resource_id(Text) : The ID of the ssh public key to be deleted.

    Request Syntax:
        [iam-user-ssh-key-id]:
          aws.iam.user_ssh_key.absent:
              - name: 'string'
              - user_name: 'string'
              - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

        APKASYLXZO1EQCR0ZAL7:
          aws.iam.user_ssh_key.absent:
          - user_name:  my-iam-user-1
          - name: APKASYLXZO1EQCR0ZAL7
          - resource_id: APKASYLXZO1EQCR0ZAL7
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before_ret = await hub.exec.aws.iam.user_ssh_key.get_ssh_key_if_attached_to_user(
        ctx, user_name=user_name, ssh_public_key_id=resource_id
    )
    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.user_ssh_key", name=name
        )
    elif ctx.get("test", False):
        ret_ssh_public_key_body = await hub.exec.boto3.client.iam.get_ssh_public_key(
            ctx=ctx,
            UserName=user_name,
            SSHPublicKeyId=resource_id,
            Encoding="SSH",
        )
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_ssh_key_to_present(
            raw_resource=before_ret["ret"],
            idem_resource_name=resource_id,
            ret_ssh_public_key_body=ret_ssh_public_key_body,
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.iam.user_ssh_key", name=name
        )
        return result
    else:
        ret_ssh_public_key_body = await hub.exec.boto3.client.iam.get_ssh_public_key(
            ctx=ctx,
            UserName=user_name,
            SSHPublicKeyId=resource_id,
            Encoding="SSH",
        )
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_ssh_key_to_present(
            raw_resource=before_ret["ret"],
            idem_resource_name=resource_id,
            ret_ssh_public_key_body=ret_ssh_public_key_body,
        )
        ret = await hub.exec.boto3.client.iam.delete_ssh_public_key(
            ctx, UserName=user_name, SSHPublicKeyId=name
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.iam.user_ssh_key", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function
    Lists all ssh public keys  of all IAM users.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.user_ssh_key
    """

    result = {}
    # Fetch all users and then ssh public keys for each of them
    users = await hub.exec.boto3.client.iam.list_users(ctx)
    if not users["result"]:
        hub.log.debug(f"Could not describe user {users['comment']}")
        return {}
    user_names = [user.get("UserName") for user in users["ret"]["Users"]]
    for user_name in user_names:
        ret_attached_ssh_public_keys = (
            await hub.exec.boto3.client.iam.list_ssh_public_keys(
                ctx=ctx, UserName=user_name
            )
        )
        if not ret_attached_ssh_public_keys["result"]:
            hub.log.warning(
                f"Could not get attached ssh public keys with user {user_name} with error"
                f" {ret_attached_ssh_public_keys['comment']} . Describe will skip this user and continue."
            )
        else:
            ssh_public_keys_list = ret_attached_ssh_public_keys["ret"].get(
                "SSHPublicKeys"
            )
            if ssh_public_keys_list:
                for resource in ssh_public_keys_list:
                    resource_id = resource.get("SSHPublicKeyId")
                    # get ssh key body
                    ret_ssh_public_key_body = (
                        await hub.exec.boto3.client.iam.get_ssh_public_key(
                            ctx=ctx,
                            UserName=user_name,
                            SSHPublicKeyId=resource_id,
                            Encoding="SSH",
                        )
                    )
                    resource_translated = hub.tool.aws.iam.conversion_utils.convert_raw_user_ssh_key_to_present(
                        raw_resource=resource,
                        idem_resource_name=resource_id,
                        ret_ssh_public_key_body=ret_ssh_public_key_body,
                    )

                    result[resource_id] = {
                        "aws.iam.user_ssh_key.present": [
                            {parameter_key: parameter_value}
                            for parameter_key, parameter_value in resource_translated.items()
                        ]
                    }
    return result
